@@ -4,6 +4,8 @@
 from __future__ import print_function
 
 import os
+import ssl
+
 import boto3
 import contextlib
 import io
@@ -17,6 +19,7 @@ from uuid import uuid4
 
 import IPython
 import botocore
+import requests
 from botocore.config import Config as BotocoreConfig
 import filelock
 import sparkmagic.utils.configuration as conf
@@ -36,6 +39,7 @@ from ipywidgets import widgets
 from requests_kerberos import REQUIRED
 from sagemaker_studio_sparkmagic_lib.emr import EMRCluster
 from sagemaker_studio_sparkmagic_lib.kerberos import write_krb_conf
+from sparkmagic.livyclientlib.exceptions import HttpClientException, EXPECTED_EXCEPTIONS
 
 from ..resource.emr.auth import ClusterSessionCredentialsProvider
 from ..resource.emr.auth import ClusterAuthUtils
@@ -53,6 +57,7 @@ from ..utils.constants import (
     SERVICE,
     OPERATION,
     SAGEMAKER_ANALYTICS_LOG_BASE_DIRECTORY,
+    VerifyCertificateArgument,
 )
 from ..utils.emr_constants import (
     IPYTHON_KERNEL,
@@ -67,14 +72,16 @@ from ..utils.emr_constants import (
     AUTH_TYPE_KERBEROS,
     AUTH_TYPE_BASIC_ACCESS,
     EMR_CONNECTION_LOG_FILE,
+    PUBLIC_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE,
+    SELF_SIGNED_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE,
 )
 from ..utils.logging_utils import ServiceFileLogger
-from ..utils.resource_check import check_host_and_port, is_ssl_enabled
+from ..utils.resource_check import check_host_and_port, is_ssl_enabled, verify_ssl_cert
 from ..utils.service_metrics import (
     records_service_metrics,
 )
 
-from ..utils.string_utils import is_blank
+from ..utils.string_utils import is_blank, unquote_ends
 
 # This logger emits logs to sys.stdout, which is visible in customer's notebook in Studio
 sys_out_logger = logging.getLogger(__name__)
@@ -100,16 +107,32 @@ def _magic_kernel_connect_to_emr(
 
     _handle_kerberos_endpoint_override(emr_cluster=emr_cluster)
 
-    # Pass the livy endpoint to connect to, through the internal cell magic without spark.conf
-    change_endpoint_magic = ipy.find_line_magic("_do_not_call_change_endpoint")
-    change_endpoint_magic(endpoint_magic_line)
+    existing_all_errors_are_fatal_conf = _configure_ssl_verification_pre_session_start()
+    livy_client_expected_exceptions = tuple(EXPECTED_EXCEPTIONS)
 
-    # Start spark session
-    start_session_magic = ipy.find_cell_magic("_do_not_call_start_session")
+    session_started = False
     error_message = None
-    session_started = start_session_magic("")
-    if not session_started:
-        error_message = "Failed to start spark session."
+
+    try:
+        # Pass the livy endpoint to connect to, through the internal cell magic without spark.conf
+        change_endpoint_magic = ipy.find_line_magic("_do_not_call_change_endpoint")
+        change_endpoint_magic(endpoint_magic_line)
+
+        # Start spark session
+        start_session_magic = ipy.find_cell_magic("_do_not_call_start_session")
+        session_started = start_session_magic("")
+
+        if not session_started:
+            error_message = "Failed to start spark session."
+
+    except livy_client_expected_exceptions as err:
+        error_message = _handle_livy_client_expected_exceptions(
+            err, args, existing_all_errors_are_fatal_conf
+        )
+    finally:
+        conf.override(
+            conf.all_errors_are_fatal.__name__, existing_all_errors_are_fatal_conf
+        )
 
     _echo_response_to_iopub_web_socket(
         _build_response(
@@ -122,6 +145,45 @@ def _magic_kernel_connect_to_emr(
     )
 
 
+def _configure_ssl_verification_pre_session_start():
+    existing_all_errors_are_fatal_conf = conf.all_errors_are_fatal()
+    if not conf.ignore_ssl_errors():
+        """
+        Sparkmagic swallows certain errors, preventing us from knowing if we were not able to connect due to
+        SSL cert verification error. Set all_errors_are_fatal flag to enable raising these errors.
+        https://github.com/jupyter-incubator/sparkmagic/blob/b5246e8cd0bddf6c5df5761389e4a45fb3fc3032/sparkmagic/sparkmagic/livyclientlib/exceptions.py#L99
+        """
+        conf.override(conf.all_errors_are_fatal.__name__, True)
+    return existing_all_errors_are_fatal_conf
+
+
+def _handle_livy_client_expected_exceptions(
+    err, args, existing_all_errors_are_fatal_conf
+):
+    error_message = str(err)
+    if isinstance(err, HttpClientException):
+        if (
+            args.verify_certificate.type
+            == VerifyCertificateArgument.VerifyCertificateArgumentType.PUBLIC_CA_CERT
+        ):
+            error_message = (
+                error_message + PUBLIC_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE
+            )
+        elif (
+            args.verify_certificate.type
+            == VerifyCertificateArgument.VerifyCertificateArgumentType.PATH_TO_CERT
+        ):
+            error_message = (
+                error_message + SELF_SIGNED_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE
+            )
+
+    if existing_all_errors_are_fatal_conf:
+        # Re-raise the exception with appropriate class if requested.
+        raise err.__class__(error_message)
+
+    return error_message
+
+
 def _get_endpoint_magic_line(
     emr_client, args, emr_cluster, username, password, kernel_name
 ):
@@ -129,7 +191,7 @@ def _get_endpoint_magic_line(
         args=args, emr_client=emr_client, emr_cluster=emr_cluster
     )
 
-    livy_endpoint = _prepare_livy_endpoint(emr_cluster, int(livy_port))
+    livy_endpoint = _prepare_livy_endpoint(emr_cluster, int(livy_port), args)
     if kernel_name in MAGIC_KERNELS:
         return _get_magic_kernel_endpoint_magic_line(
             args=args,
@@ -146,7 +208,7 @@ def _get_endpoint_magic_line(
     )
 
 
-def _prepare_livy_endpoint(emr_cluster, livy_port):
+def _prepare_livy_endpoint(emr_cluster, livy_port, args):
     # fast fail by check livy port availability. If fail, do not proceed to compose the endpoint.
     private_livy_host_name = emr_cluster.primary_node_private_dns_name()
     if check_host_and_port(private_livy_host_name, livy_port):
@@ -174,7 +236,32 @@ def _prepare_livy_endpoint(emr_cluster, livy_port):
     protocol = "http"
     if is_ssl_enabled(livy_host_name, livy_port):
         protocol = "https"
-        _enable_ignore_ssl_errors()
+
+        # Verify SSL Cert if asked. This will pre-emptively surface any SSLError.
+        if (
+            args.verify_certificate.type
+            != VerifyCertificateArgument.VerifyCertificateArgumentType.IGNORE_CERT_VERIFICATION
+        ):
+            exc = verify_ssl_cert(livy_host_name, livy_port, args.verify_certificate)
+            if exc is None:
+                print("Certificate verification for HTTPS succeeded.")
+            elif isinstance(exc, requests.exceptions.SSLError) or isinstance(
+                exc, ssl.SSLError
+            ):
+                error_to_message_map = {
+                    VerifyCertificateArgument.VerifyCertificateArgumentType.PATH_TO_CERT: SELF_SIGNED_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE,
+                    VerifyCertificateArgument.VerifyCertificateArgumentType.PUBLIC_CA_CERT: PUBLIC_SSL_CERT_VERIFICATION_FAILURE_ERROR_MESSAGE,
+                }
+
+                print(
+                    "Certificate verification for HTTPS request failed.{}. Error: [{}: {}] ".format(
+                        error_to_message_map[args.verify_certificate.type],
+                        exc.__class__.__name__,
+                        exc,
+                    )
+                )
+
+        _set_ssl_configs(args)
 
     livy_endpoint = "{0}://{1}:{2}".format(protocol, livy_host_name, livy_port)
     return livy_endpoint
@@ -449,18 +536,34 @@ def _ipython_kernel_connect_to_emr(
     _handle_kerberos_endpoint_override(emr_cluster=emr_cluster)
 
     ipy.run_line_magic("spark", "cleanup")
-    ipy.run_line_magic("spark", endpoint_magic_line)
 
-    # Check if session is created. Return true if sagemaker session name prefix is in '%spark info' response.
-    # Sagemaker session name prefix is under control and can be used as stable contract.
-    spark_info_output = io.StringIO()
-    with contextlib.redirect_stdout(spark_info_output):
-        ipy.run_line_magic("spark", "info")
-
-    session_started = SPARK_SESSION_NAME_PREFIX in spark_info_output.getvalue()
+    existing_all_errors_are_fatal_conf = _configure_ssl_verification_pre_session_start()
+    livy_client_expected_exceptions = tuple(EXPECTED_EXCEPTIONS)
     error_message = None
-    if not session_started:
-        error_message = "Failed to start spark session."
+
+    try:
+        ipy.run_line_magic("spark", endpoint_magic_line)
+    except livy_client_expected_exceptions as err:
+        error_message = _handle_livy_client_expected_exceptions(
+            err, args, existing_all_errors_are_fatal_conf
+        )
+    finally:
+        conf.override(
+            conf.all_errors_are_fatal.__name__, existing_all_errors_are_fatal_conf
+        )
+
+    session_started = False
+    if error_message is None:
+        # Check if session is created. Return true if sagemaker session name prefix is in '%spark info' response.
+        # Sagemaker session name prefix is under control and can be used as stable contract.
+        spark_info_output = io.StringIO()
+        with contextlib.redirect_stdout(spark_info_output):
+            ipy.run_line_magic("spark", "info")
+
+        session_started = SPARK_SESSION_NAME_PREFIX in spark_info_output.getvalue()
+
+        if not session_started:
+            error_message = "Failed to start spark session."
 
     _echo_response_to_iopub_web_socket(
         _build_response(
@@ -488,11 +591,41 @@ def _handle_kerberos_endpoint_override(emr_cluster):
         )
 
 
-def _enable_ignore_ssl_errors():
+def _set_ssl_configs(args):
     """
-    Set ignore_ssl_errors as true so the self-signed certificate can be accepted.
+    Determines how to verify ssl certificate validation. If verify_certificate is set to:
+    True => Certificate validation is performed with a public cert.
+    False => Certificate validation is not performed
+    Path-to-cert-file => Certificate validation is performed using the provided cert file.
+    :param args:
+    :return:
     """
-    conf.override(conf.ignore_ssl_errors.__name__, True)
+    if (
+        args.verify_certificate.type
+        == VerifyCertificateArgument.VerifyCertificateArgumentType.PUBLIC_CA_CERT
+    ):
+        conf.override(conf.ignore_ssl_errors.__name__, False)
+    elif (
+        args.verify_certificate.type
+        == VerifyCertificateArgument.VerifyCertificateArgumentType.IGNORE_CERT_VERIFICATION
+    ):
+        print(
+            "WARN: Skipping verification of SSL certificate as verify_certificate is set to False."
+        )
+        conf.override(conf.ignore_ssl_errors.__name__, True)
+    elif (
+        args.verify_certificate.type
+        == VerifyCertificateArgument.VerifyCertificateArgumentType.PATH_TO_CERT
+    ):
+        cert_file_path = args.verify_certificate.value
+        if os.path.isfile(cert_file_path):
+            conf.override(conf.custom_certfiles_path.__name__, cert_file_path)
+        else:
+            raise InvalidParameterError(
+                "{} path does not point to a file. Please provide a valid path. Refer to %sm_analytics? for all supported options. ".format(
+                    cert_file_path
+                )
+            )
 
 
 def _get_livy_port_from_cluster_configuration(emr_cluster):
@@ -738,6 +871,16 @@ class SagemakerAnalytics(Magics):
         default=None,
         help="The AWS Secrets Manager SecretID.",
     )
+    @argument(
+        "--verify-certificate",
+        type=str,
+        default="False",
+        help="Determine if SSL certificate should be verified when using HTTPS to connect to EMR. Supported values are {0}. "
+        "If a path-to-cert-file is provided, the certificate verification will be done with the certificate in the provided file path."
+        "Note that the default will change to TRUE in the next major version upgrade.".format(
+            VerifyCertificateArgument.VerifyCertificateArgumentType.list()
+        ),
+    )
     def sm_analytics(self, line):
         usage = "Please look at usage of %sm_analytics by executing `%sm_analytics?`."
         user_input = line
@@ -748,11 +891,24 @@ class SagemakerAnalytics(Magics):
         service = args.command[0].lower()
         operation = args.command[1].lower()
 
+        """
+        Magic argument parser, unlike python's argparse, doesn't remove quotation marks. Please sanitize input as needed.
+        https://github.com/ipython/ipython/issues/2001
+        """
+        args.verify_certificate = unquote_ends(args.verify_certificate)
+
+        args.verify_certificate = VerifyCertificateArgument(args.verify_certificate)
+
         global sagemaker_analytics_service_logger
 
         # emr
         if service == SERVICE.EMR:
             if operation == OPERATION.CONNECT:
+                print(
+                    "Note: The next major version update will change the default for `--verify-certificate` from False "
+                    "to True, enabling SSL verification for HTTPS connections by default. Refer to %sm_analytics? for "
+                    "more details about options supported by `--verify-certificate` "
+                )
                 sagemaker_analytics_service_logger = ServiceFileLogger(
                     EMR_CONNECTION_LOG_FILE,
                     SAGEMAKER_ANALYTICS_LOG_BASE_DIRECTORY,
